@@ -1,6 +1,11 @@
 /*
- * Copyright (c) 1992, Brian Berliner and Jeff Polk
- * Copyright (c) 1989-1992, Brian Berliner
+ * Copyright (C) 1986-2005 The Free Software Foundation, Inc.
+ *
+ * Portions Copyright (C) 1998-2005 Derek Price, Ximbiot <http://ximbiot.com>,
+ *                                  and others.
+ *
+ * Portions Copyright (C) 1992, Brian Berliner and Jeff Polk
+ * Portions Copyright (C) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
  * specified in the README file that comes with the CVS source distribution.
@@ -37,7 +42,7 @@ Entnode_Create (enum ent_type type, const char *user, const char *vn,
     Entnode *ent;
     
     /* Note that timestamp and options must be non-NULL */
-    ent = (Entnode *) xmalloc (sizeof (Entnode));
+    ent = xmalloc (sizeof (Entnode));
     ent->type      = type;
     ent->user      = xstrdup (user);
     ent->version   = xstrdup (vn);
@@ -83,10 +88,10 @@ write_ent_proc (Node *node, void *closure)
     if (closure != NULL && entnode->type != ENT_FILE)
 	*(int *) closure = 1;
 
-    if (fputentent(entfile, entnode))
+    if (fputentent (entfile, entnode))
 	error (1, errno, "cannot write %s", entfilename);
 
-    return (0);
+    return 0;
 }
 
 /*
@@ -157,7 +162,7 @@ Scratch_Entry (List *list, const char *fname)
 {
     Node *node;
 
-    TRACE ( 1, "Scratch_Entry(%s)", fname );
+    TRACE (TRACE_FUNCTION, "Scratch_Entry(%s)", fname);
 
     /* hashlookup to see if it is there */
     if ((node = findnode_fn (list, fname)) != NULL)
@@ -165,7 +170,7 @@ Scratch_Entry (List *list, const char *fname)
 	if (!noexec)
 	{
 	    entfilename = CVSADM_ENTLOG;
-	    entfile = open_file (entfilename, "a");
+	    entfile = xfopen (entfilename, "a");
 
 	    if (fprintf (entfile, "R ") < 0)
 		error (1, errno, "cannot write %s", entfilename);
@@ -206,10 +211,10 @@ Register (List *list, const char *fname, const char *vn, const char *ts,
     }
 #endif
 
-    TRACE ( 1, "Register(%s, %s, %s%s%s, %s, %s %s)",
-	    fname, vn, ts ? ts : "",
-	    ts_conflict ? "+" : "", ts_conflict ? ts_conflict : "",
-	    options, tag ? tag : "", date ? date : "" );
+    TRACE (TRACE_FUNCTION, "Register(%s, %s, %s%s%s, %s, %s %s)",
+	   fname, vn, ts ? ts : "",
+	   ts_conflict ? "+" : "", ts_conflict ? ts_conflict : "",
+	   options, tag ? tag : "", date ? date : "");
 
     entnode = Entnode_Create (ENT_FILE, fname, vn, ts, options, tag, date,
 			      ts_conflict);
@@ -257,7 +262,7 @@ freesdt (Node *p)
    On error, prints an error message and returns NULL.  */
 
 static Entnode *
-fgetentent(FILE *fpin, char *cmd, int *sawdir)
+fgetentent (FILE *fpin, char *cmd, int *sawdir)
 {
     Entnode *ent;
     char *line;
@@ -326,8 +331,8 @@ fgetentent(FILE *fpin, char *cmd, int *sawdir)
 	if ((cp = strchr (tag_or_date, '\n')) == NULL)
 	    continue;
 	*cp = '\0';
-	tag = (char *) NULL;
-	date = (char *) NULL;
+	tag = NULL;
+	date = NULL;
 	if (*tag_or_date == 'T')
 	    tag = tag_or_date + 1;
 	else if (*tag_or_date == 'D')
@@ -377,7 +382,7 @@ fgetentent(FILE *fpin, char *cmd, int *sawdir)
 }
 
 static int
-fputentent(FILE *fp, Entnode *p)
+fputentent (FILE *fp, Entnode *p)
 {
     switch (p->type)
     {
@@ -446,8 +451,8 @@ Entries_Open (int aflag, char *update_dir)
     ParseTag (&dirtag, &dirdate, &dirnonbranch);
     if (aflag || dirtag || dirdate)
     {
-	sdtp = (struct stickydirtag *) xmalloc (sizeof (*sdtp));
-	memset ((char *) sdtp, 0, sizeof (*sdtp));
+	sdtp = xmalloc (sizeof (*sdtp));
+	memset (sdtp, 0, sizeof (*sdtp));
 	sdtp->aflag = aflag;
 	sdtp->tag = xstrdup (dirtag);
 	sdtp->date = xstrdup (dirdate);
@@ -469,7 +474,7 @@ Entries_Open (int aflag, char *update_dir)
     }
     else
     {
-	while ((ent = fgetentent (fpin, (char *) NULL, &sawdir)) != NULL) 
+	while ((ent = fgetentent (fpin, NULL, &sawdir)) != NULL) 
 	{
 	    (void) AddEntryNode (entries, ent);
 	}
@@ -500,6 +505,7 @@ Entries_Open (int aflag, char *update_dir)
 		break;
 	    default:
 		/* Ignore unrecognized commands.  */
+		Entnode_Destroy (ent);
 	        break;
 	    }
 	}
@@ -516,8 +522,8 @@ Entries_Open (int aflag, char *update_dir)
 	sdtp->subdirs = sawdir;
     else if (! sawdir)
     {
-	sdtp = (struct stickydirtag *) xmalloc (sizeof (*sdtp));
-	memset ((char *) sdtp, 0, sizeof (*sdtp));
+	sdtp = xmalloc (sizeof (*sdtp));
+	memset (sdtp, 0, sizeof (*sdtp));
 	sdtp->subdirs = 0;
 	entries->list->data = sdtp;
 	entries->list->delproc = freesdt;
@@ -531,11 +537,11 @@ Entries_Open (int aflag, char *update_dir)
 	free (dirtag);
     if (dirdate)
 	free (dirdate);
-    return (entries);
+    return entries;
 }
 
 void
-Entries_Close(List *list)
+Entries_Close (List *list)
 {
     if (list)
     {
@@ -544,7 +550,7 @@ Entries_Close(List *list)
             if (isfile (CVSADM_ENTLOG))
 		write_entries (list);
 	}
-	dellist(&list);
+	dellist (&list);
     }
 }
 
@@ -558,7 +564,7 @@ Entries_delproc (Node *node)
 {
     Entnode *p = node->data;
 
-    Entnode_Destroy(p);
+    Entnode_Destroy (p);
 }
 
 /*
@@ -591,7 +597,7 @@ AddEntryNode (List *list, Entnode *entdata)
 
     /* put the node into the list */
     addnode (list, p);
-    return (p);
+    return p;
 }
 
 
@@ -603,7 +609,7 @@ void
 WriteTemplate (const char *update_dir, int xdotemplate, const char *repository)
 {
 #ifdef SERVER_SUPPORT
-    TRACE (1, "Write_Template (%s, %s)", update_dir, repository);
+    TRACE (TRACE_FUNCTION, "Write_Template (%s, %s)", update_dir, repository);
 
     if (noexec)
 	return;
@@ -637,17 +643,15 @@ WriteTag (const char *dir, const char *tag, const char *date, int nonbranch,
     if (noexec)
 	return;
 
-    tmp = xmalloc ((dir ? strlen (dir) : 0)
-		   + sizeof (CVSADM_TAG)
-		   + 10);
-    if (dir == NULL)
-	(void) strcpy (tmp, CVSADM_TAG);
+    if (dir != NULL)
+	tmp = Xasprintf ("%s/%s", dir, CVSADM_TAG);
     else
-	(void) sprintf (tmp, "%s/%s", dir, CVSADM_TAG);
+	tmp = xstrdup (CVSADM_TAG);
+
 
     if (tag || date)
     {
-	fout = open_file (tmp, "w+");
+	fout = xfopen (tmp, "w+");
 	if (tag)
 	{
 	    if (nonbranch)
@@ -702,9 +706,9 @@ ParseTag (char **tagp, char **datep, int *nonbranchp)
     FILE *fp;
 
     if (tagp)
-	*tagp = (char *) NULL;
+	*tagp = NULL;
     if (datep)
-	*datep = (char *) NULL;
+	*datep = NULL;
     /* Always store a value here, even in the 'D' case where the value
        is unspecified.  Shuts up tools which check for references to
        uninitialized memory.  */
@@ -821,20 +825,14 @@ subdir_record (int cmd, const char *parent, const char *dir)
     /* None of the information associated with a directory is
        currently meaningful.  */
     entnode = Entnode_Create (ENT_SUBDIR, dir, "", "", "",
-			      (char *) NULL, (char *) NULL,
-			      (char *) NULL);
+			      NULL, NULL, NULL);
 
     if (!noexec)
     {
 	if (parent == NULL)
 	    entfilename = CVSADM_ENTLOG;
 	else
-	{
-	    entfilename = xmalloc (strlen (parent)
-				   + sizeof CVSADM_ENTLOG
-				   + 10);
-	    sprintf (entfilename, "%s/%s", parent, CVSADM_ENTLOG);
-	}
+	    entfilename = Xasprintf ("%s/%s", parent, CVSADM_ENTLOG);
 
 	entfile = CVS_FOPEN (entfilename, "a");
 	if (entfile == NULL)
@@ -852,7 +850,8 @@ subdir_record (int cmd, const char *parent, const char *dir)
 	    }
 	    else
 	    {
-		sprintf (entfilename, "%s/%s", parent, CVSADM);
+		free (entfilename);
+		entfilename = Xasprintf ("%s/%s", parent, CVSADM);
 		if (! isdir (entfilename))
 		{
 		    free (entfilename);
@@ -961,7 +960,8 @@ Subdir_Deregister (List *entries, const char *parent, const char *dir)
    "cvs release" on the client side; see comment at src/release.c
    (release).  Would also allow us to stop needing Questionable.  */
 
-enum base_walk {
+enum base_walk
+{
     /* Set the revision for FILE to *REV.  */
     BASE_REGISTER,
     /* Get the revision for FILE and put it in a newly malloc'd string
@@ -994,23 +994,18 @@ base_walk (enum base_walk code, struct file_info *finfo, char **rev)
        computation probably should be broken out into a separate function,
        as recurse.c does it too and places like Entries_Open should be
        doing it.  */
-    baserev_fullname = xmalloc (sizeof (CVSADM_BASEREV)
-				+ strlen (finfo->update_dir)
-				+ 2);
-    baserev_fullname[0] = '\0';
-    baserevtmp_fullname = xmalloc (sizeof (CVSADM_BASEREVTMP)
-				   + strlen (finfo->update_dir)
-				   + 2);
-    baserevtmp_fullname[0] = '\0';
     if (finfo->update_dir[0] != '\0')
     {
-	strcat (baserev_fullname, finfo->update_dir);
-	strcat (baserev_fullname, "/");
-	strcat (baserevtmp_fullname, finfo->update_dir);
-	strcat (baserevtmp_fullname, "/");
+	baserev_fullname = Xasprintf ("%s/%s", finfo->update_dir,
+				      CVSADM_BASEREV);
+	baserevtmp_fullname = Xasprintf ("%s/%s", finfo->update_dir,
+					 CVSADM_BASEREVTMP);
+    }
+    else
+    {
+	baserev_fullname = xstrdup (CVSADM_BASEREV);
+	baserevtmp_fullname = xstrdup (CVSADM_BASEREVTMP);
     }
-    strcat (baserev_fullname, CVSADM_BASEREV);
-    strcat (baserevtmp_fullname, CVSADM_BASEREVTMP);
 
     fp = CVS_FOPEN (CVSADM_BASEREV, "r");
     if (fp == NULL)
