@@ -1,6 +1,11 @@
 /*
- * Copyright (c) 1992, Brian Berliner and Jeff Polk
- * Copyright (c) 1989-1992, Brian Berliner
+ * Copyright (C) 1986-2005 The Free Software Foundation, Inc.
+ *
+ * Portions Copyright (C) 1998-2005 Derek Price, Ximbiot <http://ximbiot.com>,
+ *                                  and others.
+ *
+ * Portions Copyright (C) 1992, Brian Berliner and Jeff Polk
+ * Portions Copyright (C) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
  * specified in the README file that comes with the CVS source distribution.
@@ -17,7 +22,7 @@
  */
 
 #include "cvs.h"
-#include "savecwd.h"
+#include "save-cwd.h"
 
 static char *get_comment (const char *user);
 static int add_rev (char *message, RCSNode *rcs, char *vfile,
@@ -31,6 +36,12 @@ static int process_import_file (char *message, char *vfile, char *vtag,
 				int targc, char *targv[]);
 static int update_rcs_file (char *message, char *vfile, char *vtag, int targc,
 			    char *targv[], int inattic);
+#ifdef PRESERVE_PERMISSIONS_SUPPORT
+static int preserve_initial_permissions (FILE *fprcs, const char *userfile,
+					 mode_t file_type, struct stat *sbp);
+#endif
+static int expand_and_copy_contents (FILE *fprcs, mode_t file_type,
+				     const char *user, FILE *fpuser);
 static void add_log (int ch, char *fname);
 
 static int repos_len;
@@ -41,12 +52,14 @@ static char *repository;
 static int conflicts;
 static int use_file_modtime;
 static char *keyword_opt = NULL;
+static bool killnew;
 
 static const char *const import_usage[] =
 {
-    "Usage: %s %s [-d] [-k subst] [-I ign] [-m msg] [-b branch]\n",
+    "Usage: %s %s [-dX] [-k subst] [-I ign] [-m msg] [-b branch]\n",
     "    [-W spec] repository vendor-tag release-tags...\n",
     "\t-d\tUse the file's modification time as the time of import.\n",
+    "\t-X\tWhen importing new files, mark their trunk revisions as dead.\n",
     "\t-k sub\tSet default RCS keyword substitution mode.\n",
     "\t-I ign\tMore files to ignore (! to reset).\n",
     "\t-b bra\tVendor branch id.\n",
@@ -70,12 +83,22 @@ import (int argc, char **argv)
     if (argc == -1)
 	usage (import_usage);
 
+    /* Force -X behaviour or not based on the CVS repository
+       CVSROOT/config setting.  */
+#ifdef CLIENT_SUPPORT
+    killnew = !current_parsed_root->isremote
+	      && config->ImportNewFilesToVendorBranchOnly;
+#else /* !CLIENT_SUPPORT */
+    killnew = config->ImportNewFilesToVendorBranchOnly;
+#endif /* CLIENT_SUPPORT */
+
+
     ign_setup ();
     wrap_setup ();
 
     vbranch = xstrdup (CVSBRANCH);
     optind = 0;
-    while ((c = getopt (argc, argv, "+Qqdb:m:I:k:W:")) != -1)
+    while ((c = getopt (argc, argv, "+Qqdb:m:I:k:W:X")) != -1)
     {
 	switch (c)
 	{
@@ -114,7 +137,8 @@ import (int argc, char **argv)
 #else
 		use_editor = 0;
 #endif
-		message = xstrdup(optarg);
+		if (message) free (message);
+		message = xstrdup (optarg);
 		break;
 	    case 'I':
 		ign_add (optarg, 0);
@@ -129,6 +153,9 @@ import (int argc, char **argv)
 	    case 'W':
 		wrap_add (optarg, 0);
 		break;
+	    case 'X':
+		killnew = true;
+		break;
 	    case '?':
 	    default:
 		usage (import_usage);
@@ -159,9 +186,9 @@ import (int argc, char **argv)
      * Could abstract this to valid_module_path, but I don't think we'll need
      * to call it from anywhere else.
      */
-    if ((cp = strstr(argv[0], "CVS")) &&   /* path contains "CVS" AND ... */
-        ((cp == argv[0]) || ISSLASH(*(cp-1))) && /* /^CVS/ OR m#/CVS# AND ... */
-        ((*(cp+3) == '\0') || ISSLASH(*(cp+3))) /* /CVS$/ OR m#CVS/# */
+    if ((cp = strstr (argv[0], "CVS")) &&   /* path contains "CVS" AND ... */
+        ((cp == argv[0]) || ISSLASH (*(cp-1))) && /* /^CVS/ OR m#/CVS# AND ... */
+        ((*(cp+3) == '\0') || ISSLASH (*(cp+3))) /* /CVS$/ OR m#CVS/# */
        )
     {
         error (0, 0,
@@ -179,29 +206,21 @@ import (int argc, char **argv)
 		error (1, 0, "tag `%s' was specified more than once", argv[i]);
     }
 
-    /* XXX - this should be a module, not just a pathname */
-    if (!isabsolute (argv[0]) && pathname_levels (argv[0]) == 0)
-    {
-	if (current_parsed_root == NULL)
-	{
-	    error (0, 0, "missing CVSROOT environment variable\n");
-	    error (1, 0, "Set it or specify the '-d' option to %s.",
-		   program_name);
-	}
-	repository = xmalloc (strlen (current_parsed_root->directory)
-			      + strlen (argv[0])
-			      + 2);
-	(void) sprintf (repository, "%s/%s", current_parsed_root->directory, argv[0]);
-	repos_len = strlen (current_parsed_root->directory);
-    }
-    else
-    {
+    if (isabsolute (argv[0]) || pathname_levels (argv[0]) > 0)
 	/* It is somewhere between a security hole and "unexpected" to
 	   let the client start mucking around outside the cvsroot
 	   (wouldn't get the right CVSROOT configuration, &c).  */
 	error (1, 0, "directory %s not relative within the repository",
 	       argv[0]);
+
+    if (current_parsed_root == NULL)
+    {
+	error (0, 0, "missing CVSROOT environment variable\n");
+	error (1, 0, "Set it or specify the '-d' option to %s.",
+	       program_name);
     }
+    repository = Xasprintf ("%s/%s", current_parsed_root->directory, argv[0]);
+    repos_len = strlen (current_parsed_root->directory);
 
     /*
      * Consistency checks on the specified vendor branch.  It must be
@@ -209,11 +228,22 @@ import (int argc, char **argv)
      * support branching to a single level, so the specified vendor branch
      * must only have two dots in it (like "1.1.1").
      */
-    for (cp = vbranch; *cp != '\0'; cp++)
-	if (!isdigit ((unsigned char) *cp) && *cp != '.')
-	    error (1, 0, "%s is not a numeric branch", vbranch);
-    if (numdots (vbranch) != 2)
-	error (1, 0, "Only branches with two dots are supported: %s", vbranch);
+    {
+	regex_t pat;
+	int ret = regcomp (&pat, "^[1-9][0-9]*\\.[1-9][0-9]*\\.[1-9][0-9]*$",
+			   REG_EXTENDED);
+	assert (!ret);
+	if (regexec (&pat, vbranch, 0, NULL, 0))
+	{
+	    error (1, 0,
+"Only numeric branch specifications with two dots are\n"
+"supported by import, not `%s'.  For example: `1.1.1'.",
+		   vbranch);
+	}
+	regfree (&pat);
+    }
+
+    /* Set vhead to the branch's parent.  */
     vhead = xstrdup (vbranch);
     cp = strrchr (vhead, '.');
     *cp = '\0';
@@ -233,12 +263,12 @@ import (int argc, char **argv)
 #endif
         use_editor)
     {
-	do_editor ((char *) NULL, &message,
+	do_editor (NULL, &message,
 #ifdef CLIENT_SUPPORT
-		   current_parsed_root->isremote ? (char *) NULL :
+		   current_parsed_root->isremote ? NULL :
 #endif
 			repository,
-		   (List *) NULL);
+		   NULL);
     }
     do_verify (&message, repository);
     msglen = message == NULL ? 0 : strlen (message);
@@ -265,6 +295,8 @@ import (int argc, char **argv)
 	option_with_arg ("-m", message ? message : "");
 	if (keyword_opt != NULL)
 	    option_with_arg ("-k", keyword_opt);
+	if (killnew)
+	    send_arg ("-X");
 	/* The only ignore processing which takes place on the server side
 	   is the CVSROOT/cvsignore file.  But if the user specified -I !,
 	   the documented behavior is to not process said file.  */
@@ -296,7 +328,7 @@ import (int argc, char **argv)
     }
 #endif
 
-    if (!safe_location ( NULL ))
+    if (!safe_location (NULL))
     {
 	error (1, 0, "attempt to import the repository");
     }
@@ -324,7 +356,7 @@ import (int argc, char **argv)
 
     /* Just Do It.  */
     err = import_descend (message, argv[1], argc - 2, argv + 2);
-    if (conflicts)
+    if (conflicts || killnew)
     {
 	if (!really_quiet)
 	{
@@ -332,7 +364,10 @@ import (int argc, char **argv)
 
 	    cvs_output_tagged ("+importmergecmd", NULL);
 	    cvs_output_tagged ("newline", NULL);
-	    sprintf (buf, "%d", conflicts);
+	    if (conflicts)
+	        sprintf (buf, "%d", conflicts);
+	    else
+	        strcpy (buf, "No");
 	    cvs_output_tagged ("conflicts", buf);
 	    cvs_output_tagged ("text", " conflicts created by this import.");
 	    cvs_output_tagged ("newline", NULL);
@@ -363,8 +398,11 @@ import (int argc, char **argv)
            report any required -d option.  There is no particularly
            clean way to tell the server about the -d option used by
            the client.  */
-	(void) fprintf (logfp, "\n%d conflicts created by this import.\n",
-			conflicts);
+	if (conflicts)
+	    (void) fprintf (logfp, "\n%d", conflicts);
+	else
+	    (void) fprintf (logfp, "\nNo");
+	(void) fprintf (logfp, " conflicts created by this import.\n");
 	(void) fprintf (logfp,
 			"Use the following command to help the merge:\n\n");
 	(void) fprintf (logfp, "\t%s checkout ", program_name);
@@ -386,7 +424,7 @@ import (int argc, char **argv)
     p->type = UPDATE;
     p->delproc = update_delproc;
     p->key = xstrdup ("- Imported sources");
-    li = (struct logfile_info *) xmalloc (sizeof (struct logfile_info));
+    li = xmalloc (sizeof (struct logfile_info));
     li->type = T_TITLE;
     li->tag = xstrdup (vbranch);
     li->rev_old = li->rev_new = NULL;
@@ -409,7 +447,7 @@ import (int argc, char **argv)
     free (vbranch);
     free (vhead);
 
-    return (err);
+    return err;
 }
 
 /* Process all the files in ".", then descend into other directories.
@@ -465,11 +503,11 @@ import_descend (char *message, char *vtag, int targc, char **targv)
 		Node *n;
 
 		if (dirlist == NULL)
-		    dirlist = getlist();
+		    dirlist = getlist ();
 
-		n = getnode();
+		n = getnode ();
 		n->key = xstrdup (dp->d_name);
-		addnode(dirlist, n);
+		addnode (dirlist, n);
 	    }
 	    else if (
 #ifdef DT_DIR
@@ -519,24 +557,23 @@ import_descend (char *message, char *vtag, int targc, char **targv)
 	    err += import_descend_dir (message, p->key, vtag, targc, targv);
 	}
 
-	dellist(&dirlist);
+	dellist (&dirlist);
     }
 
-    return (err);
+    return err;
 }
 
 /*
  * Process the argument import file.
  */
 static int
-process_import_file (char *message, char *vfile, char *vtag, int targc, char **targv)
+process_import_file (char *message, char *vfile, char *vtag, int targc,
+		     char **targv)
 {
     char *rcs;
     int inattic = 0;
 
-    rcs = xmalloc (strlen (repository) + strlen (vfile) + sizeof (RCSEXT)
-		   + 5);
-    (void) sprintf (rcs, "%s/%s%s", repository, vfile, RCSEXT);
+    rcs = Xasprintf ("%s/%s%s", repository, vfile, RCSEXT);
     if (!isfile (rcs))
     {
 	char *attic_name;
@@ -551,7 +588,27 @@ process_import_file (char *message, char *vfile, char *vtag, int targc, char **t
 	    char *free_opt = NULL;
 	    char *our_opt = keyword_opt;
 
-	    free (attic_name);
+	    /* If marking newly-imported files as dead, they must be
+	       created in the attic!  */
+	    if (!killnew)
+	        free (attic_name);
+	    else 
+	    {
+		free (rcs);
+		rcs = attic_name;
+
+		/* Attempt to make the Attic directory, in case it
+		   does not exist.  */
+		(void) sprintf (rcs, "%s/%s", repository, CVSATTIC);
+		if (CVS_MKDIR (rcs, 0777 ) != 0 && errno != EEXIST)
+		    error (1, errno, "cannot make directory `%s'", rcs);
+
+		/* Note that the above clobbered the path name, so we
+		   recreate it here.  */
+		(void) sprintf (rcs, "%s/%s/%s%s", repository, CVSATTIC,
+				vfile, RCSEXT);
+	    }
+
 	    /*
 	     * A new import source file; it doesn't exist as a ,v within the
 	     * repository nor in the Attic -- create it anew.
@@ -591,7 +648,7 @@ process_import_file (char *message, char *vfile, char *vtag, int targc, char **t
 
 	    retval = add_rcs_file (message, rcs, vfile, vhead, our_opt,
 				   vbranch, vtag, targc, targv,
-				   NULL, 0, logfp);
+				   NULL, 0, logfp, killnew);
 	    if (free_opt != NULL)
 		free (free_opt);
 	    free (rcs);
@@ -605,7 +662,7 @@ process_import_file (char *message, char *vfile, char *vtag, int targc, char **t
     /*
      * an rcs file exists. have to do things the official, slow, way.
      */
-    return (update_rcs_file (message, vfile, vtag, targc, targv, inattic));
+    return update_rcs_file (message, vfile, vtag, targc, targv, inattic);
 }
 
 /*
@@ -613,7 +670,8 @@ process_import_file (char *message, char *vfile, char *vtag, int targc, char **t
  * (possibly already existing) vendor branch.
  */
 static int
-update_rcs_file (char *message, char *vfile, char *vtag, int targc, char **targv, int inattic)
+update_rcs_file (char *message, char *vfile, char *vtag, int targc,
+		 char **targv, int inattic)
 {
     Vers_TS *vers;
     int letter;
@@ -629,10 +687,9 @@ update_rcs_file (char *message, char *vfile, char *vtag, int targc, char **targv
     finfo.repository = repository;
     finfo.entries = NULL;
     finfo.rcs = NULL;
-    vers = Version_TS (&finfo, (char *) NULL, vbranch, (char *) NULL,
-		       1, 0);
+    vers = Version_TS (&finfo, NULL, vbranch, NULL, 1, 0);
     if (vers->vn_rcs != NULL
-	&& !RCS_isdead(vers->srcfile, vers->vn_rcs))
+	&& !RCS_isdead (vers->srcfile, vers->vn_rcs))
     {
 	int different;
 
@@ -648,10 +705,10 @@ update_rcs_file (char *message, char *vfile, char *vtag, int targc, char **targv
 	tocvsPath = wrap_tocvs_process_file (vfile);
 	/* FIXME: Why don't we pass tocvsPath to RCS_cmp_file if it is
            not NULL?  */
-	expand = vers->srcfile->expand != NULL &&
-			vers->srcfile->expand[0] == 'b' ? "-kb" : "-ko";
-	different = RCS_cmp_file( vers->srcfile, vers->vn_rcs, (char **)NULL,
-	                          (char *)NULL, expand, vfile );
+	expand = (vers->srcfile->expand != NULL
+		  && vers->srcfile->expand[0] == 'b') ? "-kb" : "-ko";
+	different = RCS_cmp_file (vers->srcfile, vers->vn_rcs, NULL,
+	                          NULL, expand, vfile);
 	if (tocvsPath)
 	    if (unlink_file_dir (tocvsPath) < 0)
 		error (0, errno, "cannot remove %s", tocvsPath);
@@ -669,7 +726,7 @@ update_rcs_file (char *message, char *vfile, char *vtag, int targc, char **targv
 		retval = 1;
 	    add_log ('U', vfile);
 	    freevers_ts (&vers);
-	    return (retval);
+	    return retval;
 	}
     }
 
@@ -679,7 +736,7 @@ update_rcs_file (char *message, char *vfile, char *vtag, int targc, char **targv
 	add_tags (vers->srcfile, vfile, vtag, targc, targv))
     {
 	freevers_ts (&vers);
-	return (1);
+	return 1;
     }
 
     if (vers->srcfile->branch == NULL || inattic ||
@@ -693,7 +750,7 @@ update_rcs_file (char *message, char *vfile, char *vtag, int targc, char **targv
     add_log (letter, vfile);
 
     freevers_ts (&vers);
-    return (0);
+    return 0;
 }
 
 /*
@@ -706,7 +763,7 @@ add_rev (char *message, RCSNode *rcs, char *vfile, char *vers)
     char *tocvsPath;
 
     if (noexec)
-	return (0);
+	return 0;
 
     locked = 0;
     if (vers != NULL)
@@ -725,8 +782,8 @@ add_rev (char *message, RCSNode *rcs, char *vfile, char *vers)
     }
     tocvsPath = wrap_tocvs_process_file (vfile);
 
-    status = RCS_checkin (rcs, tocvsPath == NULL ? vfile : tocvsPath,
-			  message, vbranch,
+    status = RCS_checkin (rcs, NULL, tocvsPath == NULL ? vfile : tocvsPath,
+			  message, vbranch, 0,
 			  (RCS_FLAGS_QUIET | RCS_FLAGS_KEEPFILE
 			   | (use_file_modtime ? RCS_FLAGS_MODTIME : 0)));
     ierrno = errno;
@@ -745,12 +802,12 @@ add_rev (char *message, RCSNode *rcs, char *vfile, char *vers)
 	}
 	if (locked)
 	{
-	    (void) RCS_unlock(rcs, vbranch, 0);
+	    (void) RCS_unlock (rcs, vbranch, 0);
 	    RCS_rewrite (rcs, NULL, NULL);
 	}
-	return (1);
+	return 1;
     }
-    return (0);
+    return 0;
 }
 
 /*
@@ -768,16 +825,16 @@ add_tags (RCSNode *rcs, char *vfile, char *vtag, int targc, char **targv)
     struct file_info finfo;
 
     if (noexec)
-	return (0);
+	return 0;
 
-    if ((retcode = RCS_settag(rcs, vtag, vbranch)) != 0)
+    if ((retcode = RCS_settag (rcs, vtag, vbranch)) != 0)
     {
 	ierrno = errno;
 	fperrmsg (logfp, 0, retcode == -1 ? ierrno : 0,
 		  "ERROR: Failed to set tag %s in %s", vtag, rcs->path);
 	error (0, retcode == -1 ? ierrno : 0,
 	       "ERROR: Failed to set tag %s in %s", vtag, rcs->path);
-	return (1);
+	return 1;
     }
     RCS_rewrite (rcs, NULL, NULL);
 
@@ -806,7 +863,7 @@ add_tags (RCSNode *rcs, char *vfile, char *vtag, int targc, char **targv)
 	}
     }
     freevers_ts (&vers);
-    return (0);
+    return 0;
 }
 
 /*
@@ -973,7 +1030,9 @@ get_comment (const char *user)
  *
  * INPUTS
  *   message    Log message for the addition.  Not used if add_vhead == NULL.
- *   rcs        Filename of the RCS file to create.
+ *   rcs        Filename of the RCS file to create.  Note that if 'do_killnew'
+ *		is set, this file should be in the Attic directory, and the
+ *		Attic directory must already exist.
  *   user       Filename of the file to serve as the contents of the initial
  *              revision.  Even if add_vhead is NULL, we use this to determine
  *              the modes to give the new RCS file.
@@ -994,6 +1053,8 @@ get_comment (const char *user)
  *   desclen    The number of bytes in desctext.
  *   add_logfp  Write errors to here as well as via error (), or NULL if we
  *              should use only error ().
+ *   do_killnew	Mark newly-imported files as being dead on the trunk, i.e.,
+ *		as being imported only to the vendor branch.
  *
  * RETURNS
  *   Return value is 0 for success, or nonzero for failure (in which
@@ -1004,7 +1065,7 @@ add_rcs_file (const char *message, const char *rcs, const char *user,
               const char *add_vhead, const char *key_opt,
               const char *add_vbranch, const char *vtag, int targc,
               char **targv, const char *desctext, size_t desclen,
-              FILE *add_logfp)
+              FILE *add_logfp, bool do_killnew)
 {
     FILE *fprcs, *fpuser;
     struct stat sb;
@@ -1018,9 +1079,38 @@ add_rcs_file (const char *message, const char *rcs, const char *user,
     const char *userfile;
     char *free_opt = NULL;
     mode_t file_type;
+    char *dead_revision = NULL;
 
     if (noexec)
-	return (0);
+	return 0;
+
+    if (do_killnew)
+    {
+	char *last_place;
+	int last_number;
+
+	/* If we are marking the newly imported file as dead, we must
+	   have a head revision.  */
+	if (add_vhead == NULL)
+	    error (1, 0, "killing new file attempted when no head revision is being added");
+
+	/* One extra byte for NUL, plus one for carry generated by adding
+	   one to the last number in the add_vhead revision.  */
+	dead_revision = xmalloc (strlen (add_vhead) + 2);
+	strcpy (dead_revision, add_vhead);
+
+	/* Find the loacation of the last number, which we will increment
+	   and overwrite.  Note that this handles single numbers (w/o
+	   dots), which is probably unnecessary.  */
+	if ((last_place = strrchr (dead_revision, '.')) != NULL)
+	    last_place++;
+	else
+	    last_place = dead_revision;
+	last_number = atoi (last_place);
+	if (++last_number <= 0)
+	  error (1, 0, "invalid revision number %s", add_vhead);
+	sprintf (last_place, "%d", last_number);
+    }
 
     /* Note that as the code stands now, the -k option overrides any
        settings in wrappers (whether CVSROOT/cvswrappers, -W, or
@@ -1067,7 +1157,11 @@ add_rcs_file (const char *message, const char *rcs, const char *user,
     file_type = sb.st_mode & S_IFMT;
 
     fpuser = NULL;
-    if (!preserve_perms || file_type == S_IFREG)
+    if (
+#ifdef PRESERVE_PERMISSIONS_SUPPORT
+	!config->preserve_perms ||
+#endif /* PRESERVE_PERMISSIONS_SUPPORT */
+	file_type == S_IFREG)
     {
 	fpuser = CVS_FOPEN (userfile,
 			    ((key_opt != NULL && strcmp (key_opt, "b") == 0)
@@ -1097,7 +1191,8 @@ add_rcs_file (const char *message, const char *rcs, const char *user,
      */
     if (add_vhead != NULL)
     {
-	if (fprintf (fprcs, "head     %s;\012", add_vhead) < 0)
+	if (fprintf (fprcs, "head     %s;\012",
+	             do_killnew ? dead_revision : add_vhead) < 0)
 	    goto write_error;
     }
     else
@@ -1106,7 +1201,10 @@ add_rcs_file (const char *message, const char *rcs, const char *user,
 	    goto write_error;
     }
 
-    if (add_vbranch != NULL)
+    /* This sets the default branch.  If using the 'do_killnew' functionality,
+       where imports don't show up until merged, no default branch should
+       be set.  */
+    if (add_vbranch != NULL && ! do_killnew)
     {
 	if (fprintf (fprcs, "branch   %s;\012", add_vbranch) < 0)
 	    goto write_error;
@@ -1153,19 +1251,46 @@ add_rcs_file (const char *message, const char *rcs, const char *user,
 
     /* Write the revision(s), with the date and author and so on
        (that is "delta" rather than "deltatext" from rcsfile(5)).  */
+
+    if (use_file_modtime)
+	now = sb.st_mtime;
+    else
+	(void) time (&now);
+    ftm = gmtime (&now);
+    (void) sprintf (altdate1, DATEFORM,
+		    ftm->tm_year + (ftm->tm_year < 100 ? 0 : 1900),
+		    ftm->tm_mon + 1, ftm->tm_mday, ftm->tm_hour,
+		    ftm->tm_min, ftm->tm_sec);
+    author = getcaller ();
+
+    if (do_killnew)
+    {
+	if (fprintf (fprcs, "\012%s\012", dead_revision) < 0 ||
+	fprintf (fprcs, "date     %s;  author %s;  state %s;\012",
+		 altdate1, author, RCSDEAD) < 0)
+	goto write_error;
+
+	if (fprintf (fprcs, "branches;\012") < 0)
+	    goto write_error;
+	if (fprintf (fprcs, "next    %s;\012", add_vhead) < 0)
+	    goto write_error;
+
+	if (fprintf (fprcs, "commitid        %s;\012", global_session_id) < 0)
+	    goto write_error;
+
+#ifdef PRESERVE_PERMISSIONS_SUPPORT
+	/* Store initial permissions if necessary. */
+	if (config->preserve_perms)
+	{
+	    if (preserve_initial_permissions (fprcs, userfile,
+					      file_type, sbp))
+		goto write_error;
+	}
+#endif
+    }
+
     if (add_vhead != NULL)
     {
-	if (use_file_modtime)
-	    now = sb.st_mtime;
-	else
-	    (void) time (&now);
-	ftm = gmtime (&now);
-	(void) sprintf (altdate1, DATEFORM,
-			ftm->tm_year + (ftm->tm_year < 100 ? 0 : 1900),
-			ftm->tm_mon + 1, ftm->tm_mday, ftm->tm_hour,
-			ftm->tm_min, ftm->tm_sec);
-	author = getcaller ();
-
 	if (fprintf (fprcs, "\012%s\012", add_vhead) < 0 ||
 	fprintf (fprcs, "date     %s;  author %s;  state Exp;\012",
 		 altdate1, author) < 0)
@@ -1184,52 +1309,16 @@ add_rcs_file (const char *message, const char *rcs, const char *user,
 	if (fprintf (fprcs, "next     ;\012") < 0)
 	    goto write_error;
 
+	if (fprintf (fprcs, "commitid        %s;\012", global_session_id) < 0)
+	    goto write_error;
+
 #ifdef PRESERVE_PERMISSIONS_SUPPORT
 	/* Store initial permissions if necessary. */
-	if (preserve_perms)
+	if (config->preserve_perms)
 	{
-	    if (file_type == S_IFLNK)
-	    {
-		char *link = xreadlink (userfile);
-		if (fprintf (fprcs, "symlink\t@") < 0 ||
-		    expand_at_signs (link, strlen (link), fprcs) < 0 ||
-		    fprintf (fprcs, "@;\012") < 0)
-		    goto write_error;
-		free (link);
-	    }
-	    else
-	    {
-		if (fprintf (fprcs, "owner\t%u;\012", sb.st_uid) < 0)
-		    goto write_error;
-		if (fprintf (fprcs, "group\t%u;\012", sb.st_gid) < 0)
-		    goto write_error;
-		if (fprintf (fprcs, "permissions\t%o;\012",
-			     sb.st_mode & 07777) < 0)
-		    goto write_error;
-		switch (file_type)
-		{
-		    case S_IFREG: break;
-		    case S_IFCHR:
-		    case S_IFBLK:
-#ifdef HAVE_STRUCT_STAT_ST_RDEV
-			if (fprintf (fprcs, "special\t%s %lu;\012",
-				     (file_type == S_IFCHR
-				      ? "character"
-				      : "block"),
-				     (unsigned long) sb.st_rdev) < 0)
-			    goto write_error;
-#else
-			error (0, 0,
-"can't import %s: unable to import device files on this system",
-userfile);
-#endif
-			break;
-		    default:
-			error (0, 0,
-			       "can't import %s: unknown kind of special file",
-			       userfile);
-		}
-	    }
+	    if (preserve_initial_permissions (fprcs, userfile,
+					      file_type, sbp))
+		goto write_error;
 	}
 #endif
 
@@ -1239,54 +1328,17 @@ userfile);
 		fprintf (fprcs, "date     %s;  author %s;  state Exp;\012",
 			 altdate1, author) < 0 ||
 		fprintf (fprcs, "branches ;\012") < 0 ||
-		fprintf (fprcs, "next     ;\012") < 0)
+		fprintf (fprcs, "next     ;\012") < 0 ||
+	        fprintf (fprcs, "commitid        %s;\012", global_session_id) < 0)
 		goto write_error;
 
 #ifdef PRESERVE_PERMISSIONS_SUPPORT
 	    /* Store initial permissions if necessary. */
-	    if (preserve_perms)
+	    if (config->preserve_perms)
 	    {
-		if (file_type == S_IFLNK)
-		{
-		    char *link = xreadlink (userfile);
-		    if (fprintf (fprcs, "symlink\t@") < 0 ||
-			expand_at_signs (link, strlen (link), fprcs) < 0 ||
-			fprintf (fprcs, "@;\012") < 0)
-			goto write_error;
-		    free (link);
-		}
-		else
-		{
-		    if (fprintf (fprcs, "owner\t%u;\012", sb.st_uid) < 0 ||
-			fprintf (fprcs, "group\t%u;\012", sb.st_gid) < 0 ||
-			fprintf (fprcs, "permissions\t%o;\012",
-				 sb.st_mode & 07777) < 0)
-			goto write_error;
-	    
-		    switch (file_type)
-		    {
-			case S_IFREG: break;
-			case S_IFCHR:
-			case S_IFBLK:
-#ifdef HAVE_STRUCT_STAT_ST_RDEV
-			    if (fprintf (fprcs, "special\t%s %lu;\012",
-					 (file_type == S_IFCHR
-					  ? "character"
-					  : "block"),
-					 (unsigned long) sb.st_rdev) < 0)
-				goto write_error;
-#else
-			    error (0, 0,
-"can't import %s: unable to import device files on this system",
-userfile);
-#endif
-			    break;
-			default:
-			    error (0, 0,
-			      "cannot import %s: special file of unknown type",
-			       userfile);
-		    }
-		}
+		if (preserve_initial_permissions (fprcs, userfile,
+						  file_type, sbp))
+		    goto write_error;
 	    }
 #endif
 
@@ -1312,6 +1364,29 @@ userfile);
 
     /* Now write the log messages and contents for the revision(s) (that
        is, "deltatext" rather than "delta" from rcsfile(5)).  */
+
+    if (do_killnew)
+    {
+	if (fprintf (fprcs, "\012%s\012", dead_revision) < 0 ||
+	    fprintf (fprcs, "log\012@") < 0)
+	    goto write_error;
+	if (fprintf (fprcs, "Revision %s was added on the vendor branch.\012",
+		     add_vhead) < 0)
+	    goto write_error;
+	if (fprintf (fprcs, "@\012") < 0 ||
+	    fprintf (fprcs, "text\012@") < 0)
+	{
+	    goto write_error;
+	}
+
+	/* Now copy over the contents of the file, expanding at signs.  */
+	if (expand_and_copy_contents (fprcs, file_type, user, fpuser))
+	    goto write_error;
+
+	if (fprintf (fprcs, "@\012\012") < 0)
+	    goto write_error;
+    }
+
     if (add_vhead != NULL)
     {
 	if (fprintf (fprcs, "\012%s\012", add_vhead) < 0 ||
@@ -1337,28 +1412,19 @@ userfile);
 	}
 
 	/* Now copy over the contents of the file, expanding at signs.
-	   If preserve_perms is set, do this only for regular files. */
-	if (!preserve_perms || file_type == S_IFREG)
+	 * If config->preserve_perms is set, do this only for regular files.
+	 */
+	if (!do_killnew)
 	{
-	    char buf[8192];
-	    unsigned int len;
-
-	    while (1)
-	    {
-		len = fread (buf, 1, sizeof buf, fpuser);
-		if (len == 0)
-		{
-		    if (ferror (fpuser))
-			error (1, errno, "cannot read file %s for copying",
-			       user);
-		    break;
-		}
-		if (expand_at_signs (buf, len, fprcs) < 0)
-		    goto write_error;
-	    }
+            /* Now copy over the contents of the file, expanding at signs,
+	       if not done as part of do_killnew handling above.  */
+	    if (expand_and_copy_contents (fprcs, file_type, user, fpuser))
+	        goto write_error;
 	}
+
 	if (fprintf (fprcs, "@\012\012") < 0)
 	    goto write_error;
+
 	if (add_vbranch != NULL)
 	{
 	    if (fprintf (fprcs, "\012%s.1\012", add_vbranch) < 0 ||
@@ -1408,7 +1474,7 @@ userfile);
 		error (0, errno, "cannot remove %s", tocvsPath);
     if (free_opt != NULL)
 	free (free_opt);
-    return (err);
+    return err;
 
 write_error:
     ierrno = errno;
@@ -1436,10 +1502,134 @@ read_error:
     if (free_opt != NULL)
 	free (free_opt);
 
-    return (err + 1);
+    return err + 1;
 }
 
+#ifdef PRESERVE_PERMISSIONS_SUPPORT
+/* Write file permissions and symlink information for a file being
+ * added into its RCS file.
+ *
+ * INPUTS
+ *   fprcs	FILE pointer for the (newly-created) RCS file.  Permisisons
+ *		and symlink information should be written here.
+ *   userfile	Filename of the file being added.  (Used to read symbolic
+ *		link contents, for symlinks.)
+ *   file_type	File type of userfile, extracted from sbp->st_mode.
+ *   sbp	'stat' information for userfile.
+ *
+ * RETURNS
+ *   Return value is 0 for success, or nonzero for failure (in which case
+ *   no error message has yet been printed).
+ */
+static int
+preserve_initial_permissions (fprcs, userfile, file_type, sbp)
+    FILE *fprcs;
+    const char *userfile;
+    mode_t file_type;
+    struct stat *sbp;
+{
+    if (file_type == S_IFLNK)
+    {
+	char *link = Xreadlink (userfile, sbp->st_size);
+	if (fprintf (fprcs, "symlink\t@") < 0 ||
+	    expand_at_signs (link, strlen (link), fprcs) < 0 ||
+	    fprintf (fprcs, "@;\012") < 0)
+	    goto write_error;
+	free (link);
+    }
+    else
+    {
+	if (fprintf (fprcs, "owner\t%u;\012", sbp->st_uid) < 0)
+	    goto write_error;
+	if (fprintf (fprcs, "group\t%u;\012", sbp->st_gid) < 0)
+	    goto write_error;
+	if (fprintf (fprcs, "permissions\t%o;\012",
+		     sbp->st_mode & 07777) < 0)
+	    goto write_error;
+	switch (file_type)
+	{
+	    case S_IFREG: break;
+	    case S_IFCHR:
+	    case S_IFBLK:
+#ifdef HAVE_STRUCT_STAT_ST_RDEV
+		if (fprintf (fprcs, "special\t%s %lu;\012",
+			     (file_type == S_IFCHR
+			      ? "character"
+			      : "block"),
+			     (unsigned long) sbp->st_rdev) < 0)
+		    goto write_error;
+#else
+		error (0, 0,
+"can't import %s: unable to import device files on this system",
+userfile);
+#endif
+		break;
+	    default:
+		error (0, 0,
+		       "can't import %s: unknown kind of special file",
+		       userfile);
+	}
+    }
+    return 0;
 
+write_error:
+    return 1;
+}
+#endif /* PRESERVE_PERMISSIONS_SUPPORT */
+
+/* Copy file contents into an RCS file, expanding at signs.
+ *
+ * If config->preserve_perms is set, nothing is copied if the source is not
+ * a regular file.
+ *
+ * INPUTS
+ *   fprcs	FILE pointer for the (newly-created) RCS file.  The expanded
+ *		contents should be written here.
+ *   file_type	File type of the data source.  No data is copied if
+ *		preserve_permissions is set and the source is not a
+ *		regular file.
+ *   user	Filename of the data source (used to print error messages).
+ *   fpuser	FILE pointer for the data source, whose data is being
+ *		copied into the RCS file.
+ *
+ * RETURNS
+ *   Return value is 0 for success, or nonzero for failure (in which case
+ *   no error message has yet been printed).
+ */
+static int
+expand_and_copy_contents (fprcs, file_type, user, fpuser)
+    FILE *fprcs, *fpuser;
+    mode_t file_type;
+    const char *user;
+{
+    if (
+#ifdef PRESERVE_PERMISSIONS_SUPPORT
+	!config->preserve_perms ||
+#endif /* PRESERVE_PERMISSIONS_SUPPORT */
+	file_type == S_IFREG)
+    {
+	char buf[8192];
+	unsigned int len;
+
+	while (1)
+	{
+	    len = fread (buf, 1, sizeof buf, fpuser);
+	    if (len == 0)
+	    {
+		if (ferror (fpuser))
+		    error (1, errno, "cannot read file %s for copying",
+			   user);
+		break;
+	    }
+	    if (expand_at_signs (buf, len, fprcs) < 0)
+		goto write_error;
+	}
+    }
+    return 0;
+
+write_error:
+    return 1;
+}
 
 /*
  * Write SIZE bytes at BUF to FP, expanding @ signs into double @
@@ -1447,7 +1637,7 @@ read_error:
  * to indicate the error.  If not, return a nonnegative value.
  */
 int
-expand_at_signs (const char *buf, off_t size, FILE *fp)
+expand_at_signs (const char *buf, size_t size, FILE *fp)
 {
     register const char *cp, *next;
 
@@ -1512,7 +1702,8 @@ add_log (int ch, char *fname)
  * Note that we do not follow symbolic links here, which is a feature!
  */
 static int
-import_descend_dir (char *message, char *dir, char *vtag, int targc, char **targv)
+import_descend_dir (char *message, char *dir, char *vtag, int targc,
+		    char **targv)
 {
     struct saved_cwd cwd;
     char *cp;
@@ -1520,11 +1711,11 @@ import_descend_dir (char *message, char *dir, char *vtag, int targc, char **targ
     char *rcs = NULL;
 
     if (islink (dir))
-	return (0);
+	return 0;
     if (save_cwd (&cwd))
     {
-	fperrmsg (logfp, 0, 0, "ERROR: cannot get working directory");
-	return (1);
+	fperrmsg (logfp, 0, errno, "Failed to save current directory.");
+	return 1;
     }
 
     /* Concatenate DIR to the end of REPOSITORY.  */
@@ -1536,10 +1727,7 @@ import_descend_dir (char *message, char *dir, char *vtag, int targc, char **targ
     }
     else
     {
-	char *new = xmalloc (strlen (repository) + strlen (dir) + 10);
-	strcpy (new, repository);
-	(void) strcat (new, "/");
-	(void) strcat (new, dir);
+	char *new = Xasprintf ("%s/%s", repository, dir);
 	free (repository);
 	repository = new;
     }
@@ -1551,7 +1739,7 @@ import_descend_dir (char *message, char *dir, char *vtag, int targc, char **targ
 #endif
 	error (0, 0, "Importing %s", repository);
 
-    if ( CVS_CHDIR (dir) < 0)
+    if (CVS_CHDIR (dir) < 0)
     {
 	ierrno = errno;
 	fperrmsg (logfp, 0, ierrno, "ERROR: cannot chdir to %s", repository);
@@ -1565,9 +1753,8 @@ import_descend_dir (char *message, char *dir, char *vtag, int targc, char **targ
     if (!isdir (repository))
 #endif
     {
-	rcs = xmalloc (strlen (repository) + sizeof (RCSEXT) + 5);
-	(void) sprintf (rcs, "%s%s", repository, RCSEXT);
-	if (isfile (repository) || isfile(rcs))
+	rcs = Xasprintf ("%s%s", repository, RCSEXT);
+	if (isfile (repository) || isfile (rcs))
 	{
 	    fperrmsg (logfp, 0, 0,
 		      "ERROR: %s is a file, should be a directory!",
@@ -1596,8 +1783,9 @@ import_descend_dir (char *message, char *dir, char *vtag, int targc, char **targ
 	*cp = '\0';
     else
 	repository[0] = '\0';
-    if (restore_cwd (&cwd, NULL))
-	exit (EXIT_FAILURE);
+    if (restore_cwd (&cwd))
+	error (1, errno, "Failed to restore current directory, `%s'.",
+	       cwd.name);
     free_cwd (&cwd);
-    return (err);
+    return err;
 }
