@@ -1,6 +1,11 @@
 /*
- * Copyright (c) 1992, Brian Berliner and Jeff Polk
- * Copyright (c) 1989-1992, Brian Berliner
+ * Copyright (C) 1986-2005 The Free Software Foundation, Inc.
+ *
+ * Portions Copyright (C) 1998-2005 Derek Price, Ximbiot <http://ximbiot.com>,
+ *                                  and others.
+ *
+ * Portions Copyright (C) 1992, Brian Berliner and Jeff Polk
+ * Portions Copyright (C) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
  * specified in the README file that comes with the CVS source distribution.
@@ -44,15 +49,6 @@ struct verifymsg_proc_data
     struct stat pre_stbuf;
 };
 
-/* 
- * Should the logmsg be re-read during the do_verify phase?
- * RereadLogAfterVerify=no|stat|yes
- * LOGMSG_REREAD_NEVER  - never re-read the logmsg
- * LOGMSG_REREAD_STAT   - re-read the logmsg only if it has changed
- * LOGMSG_REREAD_ALWAYS - always re-read the logmsg
- */
-int RereadLogAfterVerify = LOGMSG_REREAD_ALWAYS;
-
 /*
  * Puts a standard header on the output which is either being prepared for an
  * editor session, or being sent to a logfile program.  The modified, added,
@@ -293,14 +289,14 @@ do_editor (const char *dir, char **messagep, const char *repository,
 
     /* run the editor */
     run_setup (Editor);
-    run_arg (fname);
+    run_add_arg (fname);
     if ((retcode = run_exec (RUN_TTY, RUN_TTY, RUN_TTY,
 			     RUN_NORMAL | RUN_SIGIGNORE)) != 0)
 	error (0, retcode == -1 ? errno : 0, "warning: editor session failed");
 
     /* put the entire message back into the *messagep variable */
 
-    fp = open_file (fname, "r");
+    fp = xfopen (fname, "r");
 
     if (*messagep)
 	free (*messagep);
@@ -447,8 +443,8 @@ do_verify (char **messagep, const char *repository)
     /* Get the mod time and size of the possibly new log message
      * in always and stat modes.
      */
-    if (RereadLogAfterVerify == LOGMSG_REREAD_ALWAYS ||
-	RereadLogAfterVerify == LOGMSG_REREAD_STAT)
+    if (config->RereadLogAfterVerify == LOGMSG_REREAD_ALWAYS ||
+	config->RereadLogAfterVerify == LOGMSG_REREAD_STAT)
     {
 	if(CVS_STAT (data.fname, &post_stbuf) != 0)
 	    error (1, errno, "cannot find size of temp file %s", data.fname);
@@ -457,8 +453,8 @@ do_verify (char **messagep, const char *repository)
     /* And reread the log message in `always' mode or in `stat' mode when it's
      * changed.
      */
-    if (RereadLogAfterVerify == LOGMSG_REREAD_ALWAYS ||
-	(RereadLogAfterVerify == LOGMSG_REREAD_STAT &&
+    if (config->RereadLogAfterVerify == LOGMSG_REREAD_ALWAYS ||
+	(config->RereadLogAfterVerify == LOGMSG_REREAD_STAT &&
 	  (data.pre_stbuf.st_mtime != post_stbuf.st_mtime ||
 	    data.pre_stbuf.st_size != post_stbuf.st_size)))
     {
@@ -476,8 +472,7 @@ do_verify (char **messagep, const char *repository)
 	    char *p;
 	    FILE *fp;
 
-	    if ((fp = open_file (data.fname, "r")) == NULL)
-		error (1, errno, "cannot open temporary file %s", data.fname);
+	    fp = xfopen (data.fname, "r");
 
 	    /* On NT, we might read less than st_size bytes,
 	       but we won't read more.  So this works.  */
@@ -599,7 +594,7 @@ Update_Logfile (const char *repository, const char *xmessage, FILE *xlogfp,
 static int
 update_logfile_proc (const char *repository, const char *filter, void *closure)
 {
-    struct ulp_data *udp = (struct ulp_data *)closure;
+    struct ulp_data *udp = closure;
     TRACE (TRACE_FUNCTION, "update_logfile_proc(%s,%s)", repository, filter);
     return logfile_write (repository, filter, udp->message, udp->logfp,
                           udp->changes);
@@ -734,6 +729,8 @@ logfile_write (const char *repository, const char *filter, const char *message,
     int pipestatus;
     const char *srepos = Short_Repository (repository);
 
+    assert (repository);
+
     /* The user may specify a format string as part of the filter.
        Originally, `%s' was the only valid string.  The string that
        was substituted for it was:
@@ -785,38 +782,47 @@ logfile_write (const char *repository, const char *filter, const char *message,
        Why this duplicates the old behavior when the format string is
        `%s' is left as an exercise for the reader. */
 
-    /* %p = shortrepos
+    /* %c = cvs_cmd_name
+     * %p = shortrepos
      * %r = repository
      * %{sVv} = file name, old revision (precommit), new revision (postcommit)
      */
-    cmdline = format_cmdline(
+    /*
+     * Cast any NULL arguments as appropriate pointers as this is an
+     * stdarg function and we need to be certain the caller gets what
+     * is expected.
+     */
+    cmdline = format_cmdline (
 #ifdef SUPPORT_OLD_INFO_FMT_STRINGS
-	                      !UseNewInfoFmtStrings, srepos,
+	                      !config->UseNewInfoFmtStrings, srepos,
 #endif /* SUPPORT_OLD_INFO_FMT_STRINGS */
 	                      filter,
+	                      "c", "s", cvs_cmd_name,
+#ifdef SERVER_SUPPORT
+	                      "R", "s", referrer ? referrer->original : "NONE",
+#endif /* SERVER_SUPPORT */
 	                      "p", "s", srepos,
 	                      "r", "s", current_parsed_root->directory,
 	                      "sVv", ",", changes,
-			             logmsg_list_to_args_proc, (void *) NULL,
-	                      (char *)NULL
-	                    );
-    if( !cmdline || !strlen( cmdline ) )
+			      logmsg_list_to_args_proc, (void *) NULL,
+	                      (char *) NULL);
+    if (!cmdline || !strlen (cmdline))
     {
-	if( cmdline ) free( cmdline );
-	error( 0, 0, "logmsg proc resolved to the empty string!" );
+	if (cmdline) free (cmdline);
+	error (0, 0, "logmsg proc resolved to the empty string!");
 	return 1;
     }
 
-    if( ( pipefp = run_popen( cmdline, "w" ) ) == NULL )
+    if ((pipefp = run_popen (cmdline, "w")) == NULL)
     {
 	if (!noexec)
-	    error( 0, 0, "cannot write entry to log filter: %s", cmdline );
-	free( cmdline );
+	    error (0, 0, "cannot write entry to log filter: %s", cmdline);
+	free (cmdline);
 	return 1;
     }
     (void) fprintf (pipefp, "Update of %s\n", repository);
     (void) fprintf (pipefp, "In directory %s:", hostname);
-    cp = xgetwd ();
+    cp = xgetcwd ();
     if (cp == NULL)
 	fprintf (pipefp, "<cannot get working directory: %s>\n\n",
 		 strerror (errno));
@@ -828,12 +834,12 @@ logfile_write (const char *repository, const char *filter, const char *message,
 
     setup_tmpfile (pipefp, "", changes);
     (void) fprintf (pipefp, "Log Message:\n%s\n", (message) ? message : "");
-    if (logfp != (FILE *) 0)
+    if (logfp)
     {
 	(void) fprintf (pipefp, "Status:\n");
 	rewind (logfp);
 	while ((c = getc (logfp)) != EOF)
-	    (void) putc ((char) c, pipefp);
+	    (void) putc (c, pipefp);
     }
     free (cmdline);
     pipestatus = pclose (pipefp);
@@ -842,7 +848,7 @@ logfile_write (const char *repository, const char *filter, const char *message,
 
 
 
-/*  This routine is calld by Parse_Info.  It runs the
+/*  This routine is called by Parse_Info.  It runs the
  *  message verification script.
  */
 static int
@@ -852,7 +858,7 @@ verifymsg_proc (const char *repository, const char *script, void *closure)
 #ifdef SUPPORT_OLD_INFO_FMT_STRINGS
     char *newscript = NULL;
 #endif /* SUPPORT_OLD_INFO_FMT_STRINGS */
-    struct verifymsg_proc_data *vpd = (struct verifymsg_proc_data *) closure;
+    struct verifymsg_proc_data *vpd = closure;
     const char *srepos = Short_Repository (repository);
 
 #ifdef SUPPORT_OLD_INFO_FMT_STRINGS
@@ -863,10 +869,7 @@ verifymsg_proc (const char *repository, const char *script, void *closure)
                "    \"%s\"\n"
                "Appending default format string (\" %%l\"), but be aware that this usage is\n"
                "deprecated.", script);
-	newscript = xmalloc (strlen (script) + 4);
-	strcpy (newscript, script);
-	strcat (newscript, " %l");
-	script = newscript;
+	script = newscript = Xasprintf ("%s %%l", script);
     }
 #endif /* SUPPORT_OLD_INFO_FMT_STRINGS */
 
@@ -892,7 +895,7 @@ verifymsg_proc (const char *repository, const char *script, void *closure)
 	if (fclose (fp) == EOF)
 	    error (1, errno, "%s", vpd->fname);
 
-	if (RereadLogAfterVerify == LOGMSG_REREAD_STAT)
+	if (config->RereadLogAfterVerify == LOGMSG_REREAD_STAT)
 	{
 	    /* Remember the status of the temp file for later */
 	    if (CVS_STAT (vpd->fname, &(vpd->pre_stbuf)) != 0)
@@ -904,19 +907,28 @@ verifymsg_proc (const char *repository, const char *script, void *closure)
 	     */
 	    sleep_past (vpd->pre_stbuf.st_mtime);
 	}
-    } /* if( vpd->fname == NULL ) */
+    } /* if (vpd->fname == NULL) */
 
+    /*
+     * Cast any NULL arguments as appropriate pointers as this is an
+     * stdarg function and we need to be certain the caller gets what
+     * is expected.
+     */
     verifymsg_script = format_cmdline (
 #ifdef SUPPORT_OLD_INFO_FMT_STRINGS
-                                       0, srepos,
+                                       false, srepos,
 #endif /* SUPPORT_OLD_INFO_FMT_STRINGS */
                                        script,
+				       "c", "s", cvs_cmd_name,
+#ifdef SERVER_SUPPORT
+				       "R", "s", referrer
+				       ? referrer->original : "NONE",
+#endif /* SERVER_SUPPORT */
                                        "p", "s", srepos,
                                        "r", "s",
                                        current_parsed_root->directory,
                                        "l", "s", vpd->fname,
-                                       (char *)NULL
-                                      );
+				       (char *) NULL);
 
 #ifdef SUPPORT_OLD_INFO_FMT_STRINGS
     if (newscript) free (newscript);
@@ -944,7 +956,7 @@ verifymsg_proc (const char *repository, const char *script, void *closure)
      * reason and should care about errno or that the process we called
      * returned -1 and the value of errno is undefined.  In other words,
      * run_exec should probably be rewritten to have two return codes.  one
-     * which is it's own exit status and one which is the child process's.  So
+     * which is its own exit status and one which is the child process's.  So
      * there.  :P
      *
      * Once run_exec is returning two error codes, we should probably be
