@@ -1,5 +1,11 @@
 /*
- * Copyright (c) 1992, Brian Berliner
+ * Copyright (C) 1986-2005 The Free Software Foundation, Inc.
+ *
+ * Portions Copyright (C) 1998-2005 Derek Price, Ximbiot <http://ximbiot.com>,
+ *                                  and others.
+ *
+ * Portions Copyright (C) 1992, Brian Berliner and Jeff Polk
+ * Portions Copyright (C) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
  * specified in the README file that comes with the CVS source distribution.
@@ -32,12 +38,12 @@ mydbm_open (char *file, int flags, int mode)
     FILE *fp;
     DBM *db;
 
-    fp = CVS_FOPEN (file, (flags & O_ACCMODE) != O_RDONLY ?
-                                 FOPEN_BINARY_READWRITE : FOPEN_BINARY_READ);
+    fp = CVS_FOPEN (file, (flags & O_ACCMODE) != O_RDONLY
+			  ?  FOPEN_BINARY_READWRITE : FOPEN_BINARY_READ);
     if (fp == NULL && !(existence_error (errno) && (flags & O_CREAT)))
-	return ((DBM *) 0);
+	return NULL;
 
-    db = (DBM *) xmalloc (sizeof (*db));
+    db = xmalloc (sizeof (*db));
     db->dbm_list = getlist ();
     db->modified = 0;
     db->name = xstrdup (file);
@@ -46,17 +52,18 @@ mydbm_open (char *file, int flags, int mode)
     {
 	mydbm_load_file (fp, db->dbm_list, file);
 	if (fclose (fp) < 0)
-	    error (0, errno, "cannot close %s", file);
+	    error (0, errno, "cannot close %s",
+		   primary_root_inverse_translate (file));
     }
-    return (db);
+    return db;
 }
 
-static int write_item (Node *, void *);
+
 
 static int
 write_item (Node *node, void *data)
 {
-    FILE *fp = (FILE *)data;
+    FILE *fp = data;
     fputs (node->key, fp);
     fputs (" ", fp);
     fputs (node->data, fp);
@@ -64,6 +71,8 @@ write_item (Node *node, void *data)
     return 0;
 }
 
+
+
 void
 mydbm_close (DBM *db)
 {
@@ -73,15 +82,17 @@ mydbm_close (DBM *db)
 	fp = CVS_FOPEN (db->name, FOPEN_BINARY_WRITE);
 	if (fp == NULL)
 	    error (1, errno, "cannot write %s", db->name);
-	walklist (db->dbm_list, write_item, (void *)fp);
+	walklist (db->dbm_list, write_item, fp);
 	if (fclose (fp) < 0)
 	    error (0, errno, "cannot close %s", db->name);
     }
     free (db->name);
     dellist (&db->dbm_list);
-    free ((char *) db);
+    free (db);
 }
 
+
+
 datum
 mydbm_fetch (DBM *db, datum key)
 {
@@ -102,13 +113,15 @@ mydbm_fetch (DBM *db, datum key)
     }
     else
     {
-	val.dptr = (char *) NULL;
+	val.dptr = NULL;
 	val.dsize = 0;
     }
     free (s);
-    return (val);
+    return val;
 }
 
+
+
 datum
 mydbm_firstkey (DBM *db)
 {
@@ -124,13 +137,15 @@ mydbm_firstkey (DBM *db)
     }
     else
     {
-	key.dptr = (char *) NULL;
+	key.dptr = NULL;
 	key.dsize = 0;
     }
     db->dbm_next = p->next;
-    return (key);
+    return key;
 }
 
+
+
 datum
 mydbm_nextkey (DBM *db)
 {
@@ -146,13 +161,15 @@ mydbm_nextkey (DBM *db)
     }
     else
     {
-	key.dptr = (char *) NULL;
+	key.dptr = NULL;
 	key.dsize = 0;
     }
     db->dbm_next = p->next;
-    return (key);
+    return key;
 }
 
+
+
 /* Note: only updates the in-memory copy, which is written out at
    mydbm_close time.  Note: Also differs from DBM in that on duplication,
    it gives a warning, rather than either DBM_INSERT or DBM_REPLACE
@@ -183,11 +200,15 @@ mydbm_store (DBM *db, datum key, datum value, int flags)
     return 0;
 }
 
+
+
+/* Load a DBM file.
+ *
+ * INPUTS
+ *   filename		Used in error messages.
+ */
 static void
 mydbm_load_file (FILE *fp, List *list, char *filename)
-             
-               
-                   	/* Used in error messages. */
 {
     char *line = NULL;
     size_t line_size;
@@ -203,8 +224,7 @@ mydbm_load_file (FILE *fp, List *list, char *filename)
 
     cont = 0;
     line_num=0;
-    while( ( line_length = 
-             getdelim( &line, &line_size, '\012', fp ) ) >= 0 )
+    while ((line_length = getdelim (&line, &line_size, '\012', fp)) >= 0)
     {
 	line_num++;
 	if (line_length > 0 && line[line_length - 1] == '\012')
@@ -284,7 +304,8 @@ mydbm_load_file (FILE *fp, List *list, char *filename)
 		if (!really_quiet)
 		    error (0, 0,
 			"warning: NULL value for key `%s' at line %d of `%s'",
-			p->key, line_num, filename);
+			p->key, line_num,
+			primary_root_inverse_translate (filename));
 		freenode (p);
 		continue;
 	    }
@@ -294,14 +315,15 @@ mydbm_load_file (FILE *fp, List *list, char *filename)
 		if (!really_quiet)
 		    error (0, 0,
 			"duplicate key found for `%s' at line %d of `%s'",
-			p->key, line_num, filename);
+			p->key, line_num,
+			primary_root_inverse_translate (filename));
 		freenode (p);
 	    }
 	}
     }
     if (line_length < 0 && !feof (fp))
-	/* FIXME: should give the name of the file.  */
-	error (0, errno, "cannot read file in mydbm_load_file");
+	error (0, errno, "cannot read file `%s' in mydbm_load_file",
+	       primary_root_inverse_translate (filename));
 
     free (line);
     free (value);
