@@ -1,6 +1,11 @@
 /*
- * Copyright (c) 1992, Brian Berliner and Jeff Polk
- * Copyright (c) 1989-1992, Brian Berliner
+ * Copyright (C) 1986-2005 The Free Software Foundation, Inc.
+ *
+ * Portions Copyright (C) 1998-2005 Derek Price, Ximbiot <http://ximbiot.com>,
+ *                                  and others.
+ *
+ * Portions Copyright (C) 1992, Brian Berliner and Jeff Polk
+ * Portions Copyright (C) 1989-1992, Brian Berliner
  * 
  * You may distribute under the terms of the GNU General Public License as
  * specified in the README file that comes with the CVS source distribution.
@@ -159,6 +164,8 @@ free_names (int *pargc, char **argv)
     *pargc = 0;				/* and set it to zero when done */
 }
 
+
+
 /* Convert LINE into arguments separated by SEPCHARS.  Set *ARGC
    to the number of arguments found, and (*ARGV)[0] to the first argument,
    (*ARGV)[1] to the second, etc.  *ARGV is malloc'd and so are each of
@@ -175,21 +182,23 @@ line2argv (int *pargc, char ***argv, char *line, char *sepchars)
 
     /* Small for testing.  */
     argv_allocated = 1;
-    *argv = (char **) xmalloc (argv_allocated * sizeof (**argv));
+    *argv = xnmalloc (argv_allocated, sizeof (**argv));
 
     *pargc = 0;
-    for (cp = strtok (line, sepchars); cp; cp = strtok ((char *) NULL, sepchars))
+    for (cp = strtok (line, sepchars); cp; cp = strtok (NULL, sepchars))
     {
 	if (*pargc == argv_allocated)
 	{
 	    argv_allocated *= 2;
-	    *argv = xrealloc (*argv, argv_allocated * sizeof (**argv));
+	    *argv = xnrealloc (*argv, argv_allocated, sizeof (**argv));
 	}
 	(*argv)[*pargc] = xstrdup (cp);
 	(*pargc)++;
     }
 }
 
+
+
 /*
  * Returns the number of dots ('.') found in an RCS revision number
  */
@@ -206,6 +215,8 @@ numdots (const char *s)
     return (dots);
 }
 
+
+
 /* Compare revision numbers REV1 and REV2 by consecutive fields.
    Return negative, zero, or positive in the manner of strcmp.  The
    two revision numbers must have the same number of fields, or else
@@ -231,6 +242,8 @@ compare_revnums (const char *rev1, const char *rev2)
     return result;
 }
 
+
+
 /* Increment a revision number.  Working on the string is a bit awkward,
    but it avoid problems with integer overflow should the revision numbers
    get really big.  */
@@ -266,6 +279,8 @@ increment_revnum (const char *rev)
     return newrev;
 }
 
+
+
 /* Return the username by which the caller should be identified in
    CVS, in contexts such as the author field of RCS files, various
    logs, etc.  */
@@ -309,10 +324,7 @@ getcaller (void)
     }
     if ((pw = (struct passwd *) getpwuid (uid)) == NULL)
     {
-	char uidname[20];
-
-	(void) sprintf (uidname, "uid%lu", (unsigned long) uid);
-	cache = xstrdup (uidname);
+	cache = Xasprintf ("uid%lu", (unsigned long) uid);
 	return cache;
     }
     cache = xstrdup (pw->pw_name);
@@ -320,17 +332,20 @@ getcaller (void)
 #endif
 }
 
-#ifdef lint
-#ifndef __GNUC__
-/* ARGSUSED */
-time_t
-get_date( char *date, struct timeb *now )
-{
-    time_t foo = 0;
 
-    return (foo);
+
+#ifdef lint
+# ifndef __GNUC__
+/* ARGSUSED */
+bool
+get_date (struct timespec *result, char const *p, struct timespec const *now)
+{
+    result->tv_sec = 0;
+    result->tv_nsec = 0;
+
+    return false;
 }
-#endif
+# endif
 #endif
 
 
@@ -355,9 +370,7 @@ get_date( char *date, struct timeb *now )
  *   function should work fine, as revision X.1 must exist, due to RCS rules.
  */
 char *
-previous_rev (rcs, rev)
-    RCSNode *rcs;
-    const char *rev;
+previous_rev (RCSNode *rcs, const char *rev)
 {
     char *p;
     char *tmp = xstrdup (rev);
@@ -401,7 +414,6 @@ previous_rev (rcs, rev)
 /* Given two revisions, find their greatest common ancestor.  If the
    two input revisions exist, then rcs guarantees that the gca will
    exist.  */
-
 char *
 gca (const char *rev1, const char *rev2)
 {
@@ -480,6 +492,8 @@ gca (const char *rev1, const char *rev2)
     return retval;
 }
 
+
+
 /* Give fatal error if REV is numeric and ARGC,ARGV imply we are
    planning to operate on more than one file.  The current directory
    should be the working directory.  Note that callers assume that we
@@ -487,7 +501,6 @@ gca (const char *rev1, const char *rev2)
    '\0' at the end of the tag name and other niceties.  Right now this
    is only called from admin.c, but if people like the concept it probably
    should also be called from diff -r, update -r, get -r, and log -r.  */
-
 void
 check_numeric (const char *rev, int argc, char **argv)
 {
@@ -507,6 +520,8 @@ check_numeric (const char *rev, int argc, char **argv)
     }
 }
 
+
+
 /*
  *  Sanity checks and any required fix-up on message passed to RCS via '-m'.
  *  RCS 5.7 requires that a non-total-whitespace, non-null message be provided
@@ -643,6 +658,8 @@ out:
     return result;
 }
 
+
+
 /* Read the entire contents of the file NAME into *BUF.
    If NAME is NULL, read from stdin.  *BUF
    is a pointer returned from malloc (or NULL), pointing to *BUFSIZE
@@ -650,9 +667,9 @@ out:
    give a fatal error.  The name of the file to use in error messages
    (typically will include a directory if we have changed directory)
    is FULLNAME.  MODE is "r" for text or "rb" for binary.  */
-
 void
-get_file (const char *name, const char *fullname, const char *mode, char **buf, size_t *bufsize, size_t *len)
+get_file (const char *name, const char *fullname, const char *mode, char **buf,
+	  size_t *bufsize, size_t *len)
 {
     struct stat s;
     size_t nread;
@@ -679,7 +696,7 @@ get_file (const char *name, const char *fullname, const char *mode, char **buf,
 	/* Convert from signed to unsigned.  */
 	filesize = s.st_size;
 
-	e = open_file (name, mode);
+	e = xfopen (name, mode);
     }
 
     if (*buf == NULL || *bufsize <= filesize)
@@ -732,11 +749,11 @@ get_file (const char *name, const char *fullname, const char *mode, char **buf,
 }
 
 
+
 /* Follow a chain of symbolic links to its destination.  FILENAME
    should be a handle to a malloc'd block of memory which contains the
    beginning of the chain.  This routine will replace the contents of
    FILENAME with the destination (a real file).  */
-
 void
 resolve_symlink (char **filename)
 {
@@ -777,6 +794,8 @@ resolve_symlink (char **filename)
     }
 }
 
+
+
 /*
  * Rename a file to an appropriate backup name based on BAKPREFIX.
  * If suffix non-null, then ".<suffix>" is appended to the new name.
@@ -786,21 +805,8 @@ resolve_symlink (char **filename)
 char *
 backup_file (const char *filename, const char *suffix)
 {
-    char *backup_name;
-
-    if (suffix == NULL)
-    {
-        backup_name = xmalloc (sizeof (BAKPREFIX) + strlen (filename) + 1);
-        sprintf (backup_name, "%s%s", BAKPREFIX, filename);
-    }
-    else
-    {
-        backup_name = xmalloc (sizeof (BAKPREFIX)
-                               + strlen (filename)
-                               + strlen (suffix)
-                               + 2);  /* one for dot, one for trailing '\0' */
-        sprintf (backup_name, "%s%s.%s", BAKPREFIX, filename, suffix);
-    }
+    char *backup_name = Xasprintf ("%s%s%s%s", BAKPREFIX, filename,
+				   suffix ? "." : "", suffix ? suffix : "");
 
     if (isfile (filename))
         copy_file (filename, backup_name);
@@ -808,13 +814,14 @@ backup_file (const char *filename, const char *suffix)
     return backup_name;
 }
 
+
+
 /*
  * Copy a string into a buffer escaping any shell metacharacters.  The
  * buffer should be at least twice as long as the string.
  *
  * Returns a pointer to the terminating NUL byte in buffer.
  */
-
 char *
 shell_escape(char *buf, const char *str)
 {
@@ -922,14 +929,9 @@ char *
 cmdlinequote (char quotes, char *s)
 {
     char *quoted = cmdlineescape (quotes, s);
-    char *buf = xmalloc(strlen(quoted)+3);
+    char *buf = Xasprintf ("%c%s%c", quotes, quoted, quotes);
 
-    buf[0] = quotes;
-    buf[1] = '\0';
-    strcat (buf, quoted);
     free (quoted);
-    buf[strlen(buf)+1] = '\0';
-    buf[strlen(buf)] = quotes;
     return buf;
 }
 
@@ -1048,13 +1050,13 @@ cmdlineescape (char quotes, char *s)
  * EXAMPLE
  *    (ignoring oldway variable and srepos since those are only around while we
  *    SUPPORT_OLD_INFO_FMT_STRINGS)
- *    format_cmdline( "/cvsroot/CVSROOT/mytaginfoproc %t %o %{sVv}",
+ *    format_cmdline ("/cvsroot/CVSROOT/mytaginfoproc %t %o %{sVv}",
  *                    "t", "s", "newtag",
  *                    "o", "s", "mov",
  *                    "xG", "ld", longintwhichwontbeusedthispass,
  *                    "sVv", ",", tlist, pretag_list_to_args_proc,
- *                      (void *) mydata,
- *                    (char *)NULL);
+ *                    (void *) mydata,
+ *                    (char *) NULL);
  *
  *    would generate the following command line, assuming two files in tlist,
  *    file1 & file2, each with old versions 1.1 and new version 1.1.2.3:
@@ -1841,32 +1843,6 @@ isabsolute (filename)
 
 
 
-/*
- * void cvs_trace(int level, const char *fmt, ...)
- *
- * Print tracing information to stderr on request.  Levels are implemented
- * as with CVSNT.
- */
-void cvs_trace (int level, const char *fmt, ...)
-{
-    if (trace >= level)
-    {
-	va_list va;
-
-	va_start (va, fmt);
-#ifdef SERVER_SUPPORT
-	fprintf (stderr,"%c -> ",server_active?'S':' ');
-#else /* ! SERVER_SUPPORT */
-	fprintf (stderr,"  -> ");
-#endif
-	vfprintf (stderr, fmt, va);
-	fprintf (stderr,"\n");
-	va_end (va);
-    }
-}
-
-
-
 /* Like xstrdup (), but can handle a NULL argument.
  */
 char *
@@ -1896,7 +1872,7 @@ Xasprintf (const char *format, ...)
 
 
 
-/* Like xasmprintf(), but consider all errors fatal (may never return NULL).
+/* Like xasnprintf(), but consider all errors fatal (may never return NULL).
  */
 char *
 Xasnprintf (char *resultbuf, size_t *lengthp, const char *format, ...)
@@ -1912,3 +1888,76 @@ Xasnprintf (char *resultbuf, size_t *lengthp, const char *format, ...)
 
     return result;
 }
+
+
+
+/* Print a warning and return false if P doesn't look like a string specifying
+ * a boolean value.
+ *
+ * Sets *VAL to the parsed value when it is found to be valid.  *VAL will not
+ * be altered when false is returned.
+ *
+ * INPUTS
+ *   infopath	Where the error is reported to be from on error.  This could
+ *		be, for example, the name of the file the boolean is being read
+ *		from.
+ *   option	An option name being parsed, reported in traces and any error
+ *		message.
+ *   p		The string to actually read the option from.
+ *   val	Pointer to where to store the boolean read from P.
+ *
+ * OUTPUTS
+ *   val	TRUE/FALSE stored, as read, when there are no errors.
+ *
+ * RETURNS
+ *   true	If VAL was read.
+ *   false	On error.
+ */
+bool
+readBool (const char *infopath, const char *option, const char *p, bool *val)
+{
+    TRACE (TRACE_FLOW, "readBool (%s, %s, %s)", infopath, option, p);
+    if (!strcasecmp (p, "no") || !strcasecmp (p, "false")
+        || !strcasecmp (p, "off") || !strcmp (p, "0"))
+    {
+	TRACE (TRACE_DATA, "Read %d for %s", *val, option);
+	*val = false;
+	return true;
+    }
+    else if (!strcasecmp (p, "yes") || !strcasecmp (p, "true")
+	     || !strcasecmp (p, "on") || !strcmp (p, "1"))
+    {
+	TRACE (TRACE_DATA, "Read %d for %s", *val, option);
+	*val = true;
+	return true;
+    }
+
+    error (0, 0, "%s: unrecognized value `%s' for `%s'",
+	   infopath, p, option);
+    return false;
+}
+
+
+
+/*
+ * Open a file, exiting with a message on error.
+ *
+ * INPUTS
+ *   name	The name of the file to open.
+ *   mode	Mode to open file in, as POSIX fopen().
+ *
+ * NOTES
+ *   If you want to handle errors, just call fopen (NAME, MODE).
+ *
+ * RETURNS
+ *   The new FILE pointer.
+ */
+FILE *
+xfopen (const char *name, const char *mode)
+{
+    FILE *fp;
+
+    if (!(fp = fopen (name, mode)))
+	error (1, errno, "cannot open %s", name);
+    return fp;
+}
