@@ -10,10 +10,12 @@
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.  */
 
-int do_update PROTO((int argc, char *argv[], char *xoptions, char *xtag,
+int do_update (int argc, char *argv[], char *xoptions, char *xtag,
 	       char *xdate, int xforce, int local, int xbuild,
 	       int xaflag, int xprune, int xpipeout, int which,
-	       char *xjoin_rev1, char *xjoin_rev2, char *preload_update_dir,
-	       int xdotemplate));
-int joining PROTO((void));
-extern int isemptydir PROTO ((char *dir, int might_not_exist));
+	       char *xjoin_rev1, char *xjoin_date1,
+	       char *xjoin_rev2, char *xjoin_date2,
+	       char *preload_update_dir, int xdotemplate,
+	       char *repository);
+int joining (void);
+extern int isemptydir (const char *dir, int might_not_exist);
