@@ -1,17 +1,3 @@
-/*
- * Copyright (C) 2003-2005 The Free Software Foundation, Inc.
- *
- * This program is free software; you can redistribute it and/or modify
- * it under the terms of the GNU General Public License as published by
- * the Free Software Foundation; either version 2, or (at your option)
- * any later version.
- *
- * This program is distributed in the hope that it will be useful,
- * but WITHOUT ANY WARRANTY; without even the implied warranty of
- * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- * GNU General Public License for more details.
- */
-
 /*
  * win32.c
  * - utility functions for cvs under win32
@@ -24,10 +10,6 @@
 #include <stdio.h>
 #include <conio.h>
 
-#ifdef HAVE_PROCESS_H
-# include <process.h>
-#endif
-
 #define WIN32_LEAN_AND_MEAN
 #include <windows.h>
 
@@ -67,20 +49,37 @@ wnt_cleanup (void)
     }
 }
 
+
+
 unsigned sleep(unsigned seconds)
 {
 	Sleep(1000*seconds);
 	return 0;
 }
 
+
+
 /*
- * Sleep at least useconds microseconds.
+ * Sleep at least some number of microseconds, specified with nanosecond
+ * resolution and rounding up to the nearest microsecond.
  */
-int usleep(unsigned long useconds)
+static int
+my_usleep (const struct timespec *ts_delay)
 {
-    /* Not very accurate, but it gets the job done */
-    Sleep(useconds/1000 + (useconds%1000 ? 1 : 0));
-    return 0;
+    struct timeval tv_delay;
+    tv_delay.tv_sec = ts_delay->tv_sec;
+    tv_delay.tv_usec = ts_delay->tv_nsec / 1000
+		       + (ts_delay->tv_nsec % 1000 ? 1 : 0);
+    return select (0, (void *) 0, (void *) 0, (void *) 0, &tv_delay);
+}
+
+
+
+int
+nanosleep (const struct timespec *requested_delay,
+	   struct timespec *remaining_delay)
+{
+    return my_usleep (requested_delay);
 }
 
 
@@ -97,13 +96,11 @@ char *win32getlogin()
 }
 
 
-#ifndef HAVE_GETPID
 pid_t
-getpid (void)
+getpid ()
 {
     return (pid_t) GetCurrentProcessId();
 }
-#endif
 
 char *
 getpass (const char *prompt)
